# Низове и binaries

## Binaries

### Конструкция
```elixir
<<>>

<< 123, 23, 1 >>

<< 0b01111011, 0b00010111, 0b00000001 >> == << 123, 23, 1 >>
# true

<< 280 >>
# <<24>>
```


### Конструкция

```elixir
<< 280::size(16) >> # Същото като << 280::16 >>
# <<1, 24>>

<< 0b00000001, 0b00011000 >> == << 280::16 >>

<< 129::7 >>
# <<1::size(7)>>
```


### Конструкция

- Интересно е съхранението на числа с плаваща запетая като `binaries`.
- Винаги се представят като `binary` от 8 байта (double precision):

```elixir
<< 5.5::float >>
# <<64, 22, 0, 0, 0, 0, 0, 0>>
```


### Конкатенация

```elixir
<< 83, 79, 83 >> <> << 24, 4 >>

<< 83, 79, 83, << 24, 4 >> >>
```


### Размер

```elixir
Kernel.byte_size(<< 83, 79, 83 >>)

byte_size(<< 34::5, 23::2, 12::2 >>)
```


### Размер

```elixir
Kernel.bit_size(<< 34::5, 23::2, 12::2 >>)
```


### Sub-binaries

С функцията `Kernel.binary_part/3` можем да боравим с части от дадена `binary` структура:

```elixir
binary_part(<< 83, 79, 83, 23, 21, 12 >>, 1, 3)
# <<79, 83, 23>>

binary_part(<< 83, 79, 83, 23, 21, 12 >>, 2, -1)
binary_part(<< 83, 79, 83, 23, 21, 12 >>, 1, 6)
binary_part(<< 83, 79, 83, 23, 21, 12 >>, -1, 3)
```


### Използване в guard

Функцията `Kernel.binary_part/3` вътрешно ползва `:erlang.binary_part/3`, която е *BIF*, позволен в guard-ове:

```elixir
case <<17, 13, 14, 15>> do
  v when binary_part(v, 2, -2) == <<17, 13>> -> :ok
  _ -> :nope
end
```


Още една функция от Erlang : `:erlang.split_binary/2`, отново ползва `:erlang.binary_part/3`, но не може да е част от guard:

```elixir
:erlang.split_binary(<<4, 65, 129>>, 2)
# {<<4, 65>>, <<129>>}
```


### Pattern matching

Като всичко друго в Elixir, *binaries* също можем да съпоставяме:

```elixir
<< x, y, x >> = << 83, 79, 83 >>
x
y
```


```elixir
<< x, y, z::binary >> = << 83, 79, 83, 43, 156 >>
z
# <<83, 43, 156>>

<< x, y, z >> = << 83, 79, 83, 43, 156 >>

<< x, y, z::bitstring >> = << 83, 79, 83, 43, 156 >>
```


### Съпоставяне и числа с плаваща запетая

```elixir
<<
  sign::size(1),
  exponent::size(11),
  mantissa::size(52)
>> = << 4815.162342::float >>

sign
# 0
```


```elixir
:math.pow(-1, sign) *
  (1 + mantissa / :math.pow(2, 52)) *
  :math.pow(2, exponent - 1023)

```


### Модификатори

```elixir
<< x::5, y::bits >> = << 225 >>
x
y

<< 0b11111::5 >> = x
<< 0b111::3 >> = y
```


```elixir
<< x::binary-size(4), _::binary >> =
  << 83, 222, 0, 345, 143, 87 >>

x # 4 байта
# <<83, 222, 0, 89>>
```


### Копиране

```elixir
x = << 83, 222, 0, 89 >> # Heap binary (< 64 bytes)
y = x <> << 225, 21 >> # Refc (concat -> 2*y size or 256 => 256 bytes)
z = y <> << 125, 156 >> # Refc, but reuses the memory
u = z <> << 2, 4 >> # Refc, but reuses the memory
a = y <> << 15, 16, 19 >> # Refc (concat -> 2*a size or 256)
```


### Съпоставяне

```elixir
defmodule BPM do
  def binary_to_list(<< head, tail::binary>>) do
    [head | binary_to_list(tail)]
  end
  def binary_to_list(<<>>), do: []
end

BPM.binary_to_list("Let's there be list!")
```


### Динамично съпоставяне

- Можем да реферираме стойности, които вече сме съпоставили.

```elixir
str = String.duplicate("a", 56)              
# "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
size = byte_size(str)                            
# 56
size_value_bin = <<size::32, str::binary>>               
# <<0, 0, 0, 56, 97, 97, 97, ...>>
<<
  payload_size::32,
  payload::binary-size(payload_size)
>> = size_value_bin
# <<0, 0, 0, 56, 97, 97, 97, ...>>
payload_size
# 56
payload
# "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
```


## Низове

```elixir
<< 0b00110000 >>
# "0"
<< 0b00110001 >>
# "1"
<< 0b00110010 >>
# "2"
<< 0b00110101 >>
# "5"
<< 0b00111001 >>
# "9"
```



```elixir
?Ъ
#=> 1066

Integer.to_string(1066, 2)
# "10000101010"
```


Префиксваме с 110 и допълваме до байт с битовете от числото (110)10000

Префиксваме с 10 и допълваме до байт с битовете, които останаха (10)101010

```elixir
<< 0b11010000, 0b10101010 >>
"Ъ"
```


### Графеми и codepoint-и

```elixir
name = "Николай"
String.codepoints(name)
# ["Н", "и", "к", "о", "л", "а", "й"]

String.graphemes(name)
String.graphemes(name) == String.codepoints(name)
```


### Графеми и codepoint-и

```elixir
name2 = "Николаи\u0306"
String.codepoints(name2)
#=> ["Н", "и", "к", "о", "л", "а", "и", ̆"<не може да се представи>"]
String.graphemes(name2)
# ["Н", "и", "к", "о", "л", "а", "й"]
```


А ето как да пишем директно codepoint-и `\u`:
```elixir
Integer.to_string(1066, 16) # Codepooint of "Ъ" in hex
# "42A"

"\u042Aгъл"

# Ако не са 4 hex цифри : \u{D...}
```


Функции като `String.reverse/1`, работят правилно, защото ползват `String.graphemes/1`.

```elixir
String.reverse("Николаи\u0306")

# Аналогично:
String.graphemes(name2) |> Enum.reverse |> Enum.join("")
```


### Дължина на низ

```elixir
Kernel.bit_size "Николаи\u0306"
Kernel.byte_size "Николаи\u0306"
String.length "Николаи\u0306"
```


### Под-низове

```elixir
String.at("Искам бира!", 6)
String.at("Искам бира!", -1)
String.at("Искам бира!", 12)
```


```elixir
<< _::binary-size(11), x::utf8, _::binary >> = "Искам бира!"

x
# 1073
<< x::utf8 >>
# б
```


```elixir
<< "Искам ", x::utf8, _::binary >> = "Искам бира!"
```


```elixir
String.next_codepoint("Николаи\u0306")
# {"Н", "иколай"}

String.next_codepoint("и\u0306")
String.next_grapheme("и\u0306")
# {"й", ""}

String.next_grapheme_size("и\u0306")
# {4, ""}
String.next_grapheme_size("\u0306")
```


```elixir
poem = """
  Ще строим завод,
  огромен завод,
  със яки
        бетонни стени!
  Мъже и жени,
  народ,
  ще строим завод
  за живота!
"""

large_factory = String.slice(poem, 18..30)
String.slice(poem, 18, 13)
```


## Обхождане на низове

```elixir
str =
  (
    Enum.to_list(?a..?z) ++
    Enum.to_list(?A..?Z) ++
    Enum.to_list(?а..?я) ++
    Enum.to_list(?А..?Я) ++
    [32, 46]
  )
  |> Stream.cycle()
  |> Stream.take(2_000_000)
  |> Enum.shuffle()
  |> List.to_string()
```


```elixir
defmodule ACounterSlice do
  def count_it(str) do
    count(str, 0)
  end

  defp count("", n), do: n
  defp count(str, n) do
    the_next_n = next_n(String.starts_with?(str, "a"), n)
    count(String.slice(str, 1..-1), the_next_n)
  end

  defp next_n(true, n), do: n + 1
  defp next_n(false, n), do: n
end
```


Това е около *3.18* секунди за низ с дължина над 2_000_000

```elixir
{time, _result} = :timer.tc(ACounterSlice, :count_it, [str])
"#{time / 1000_000}s"
```


```elixir
defmodule ACounterNextGrapheme do
  def count_it(str) do
    count(str, 0)
  end

  defp count("", n), do: n
  defp count(str, n) do
    {next_grapheme, rest} = String.next_grapheme(str)
    count(rest, next_n(next_grapheme == "a", n)
    )
  end

  defp next_n(true, n), do: n + 1
  defp next_n(false, n), do: n
end
```


Добре същият отговор, но сега за *0.8* секунди

```elixir
{time, _result} = :timer.tc(ACounterNextGrapheme, :count_it, [str])
"#{time / 1000_000}s"
```


```elixir
defmodule ACounterMatch do
  def count_it(str) do
    count(str, 0)
  end

  defp count("", n), do: n

  defp count(<< c::utf8, rest::binary >>, n)
  when c == ?a do
    count(rest, n + 1)
  end

  defp count(<< _::utf8, rest::binary >>, n) do
    count(rest, n)
  end
end
```


Тази имплементация със същия низ е около *0.1* секунди.

```elixir
{time, _result} = :timer.tc(ACounterMatch, :count_it, [str])
"#{time / 1000_000}s"
```


## Списъци от codepoint-и (charlists)

```elixir
'Erlang'
[?E, ?r, ?l, ?a, ?n, ?g]

"Erlang" == 'Erlang'
```


Списъкът се принтира в кавички, само ако codepoint-ите са в ASCII-range (0-127, и могат да се принтират).

```elixir
'Ерланг'

is_list('Ерланг')

[69, 114, 108, 97, 110, 103]
[69, 114, 108, 97, 110, 103] ++ [33]

inspect([83, 79, 83], charlists: :as_lists)
```


За да виждаме списъци от числа винаги като списъци от числа в IEX:

```elixir
IEx.configure(inspect: [charlists: :as_lists])
```